import os
from typing import Dict, Optional

import boto3
import pandas as pd
from botocore.config import Config
from pandas import DataFrame as PandasDataFrame

from dump.config_utils import load_config
from dump.files.tmp_util import TemporaryFileSystem


class S3Client:
    def __init__(
        self, creds_section: str, config_filiname: str = "s3_config.ini"
    ) -> None:
        self.creds_section = creds_section

        self.config = load_config(filename=config_filiname, section=creds_section)
        self.conn_config = Config(
            s3={
                "addressing_style": "virtual",
            },
            retries={"max_attempts": 10, "mode": "standard"},
            region_name="us-east-1",
        )

        self.client = boto3.client(
            "s3",
            **self.config,
            config=self.conn_config,
        )

    def get_files_info(self, bucket_name: str, object_folder: Optional[str] = ""):
        files_dict = {}
        response = self.client.list_objects_v2(Bucket=bucket_name, Prefix=object_folder)
        if "Contents" in response:
            for obj in response["Contents"]:
                size = obj["Size"]
                files_dict[obj["Key"]] = {
                    "size_bytes": size,
                    "size_gb": size / 1024**3,
                }
        return files_dict


class S3Upload(S3Client):
    """
    To upload files to s3
    """

    def __init__(
        self,
        bucket_name: str,
        object_folder: str = "",
        creds_section: str = "test",
    ) -> None:
        super().__init__(creds_section)
        self.bucket_name = bucket_name
        self.object_folder = object_folder

    def _get_path(self, file_name: str) -> str:
        """
        creating path where to save file in s3
        """
        path = os.path.join(self.object_folder, file_name)
        return path

    def upload_to_s3(self, tmp_path_file: str, file_name: str):
        """
        tmp_path_file: str - file location that we want to save
        file_name: str - new file name that will be displayed in s3

        upload local file into s3 bucket with new name 'file_name'
        """
        path = self._get_path(file_name)
        self.client.upload_file(tmp_path_file, self.bucket_name, path)


class S3Download(S3Client):
    def __init__(
        self,
        bucket_name: str,
        object_folder: str = "",
        creds_section: str = "test",
    ) -> None:
        super().__init__(creds_section)
        self.bucket_name = bucket_name
        self.object_folder = object_folder

        self.file_system = TemporaryFileSystem()

        # requires for not download same file more than one time
        self.__downloads_history: Dict[str, str] = {}

    def _check_path_exist(self, path: str) -> bool:
        exsist_files = self.get_files_info(self.bucket_name, self.object_folder)
        return path in exsist_files

    def _get_object_path(self, file_name) -> str:
        path = os.path.join(self.object_folder, file_name)
        return path

    def download_from_s3(self, download_file_name: str) -> str:
        """
        downloading file from s3 and put the file into temporary directory
        returns path to file in temporary directory
        """
        object_path = self._get_object_path(download_file_name)
        if self._check_path_exist(object_path):
            if object_path not in self.__downloads_history:
                save_path = self.file_system.save_path()
                self.client.download_file(self.bucket_name, object_path, save_path)
                self.__downloads_history[object_path] = save_path

            return self.__downloads_history[object_path]
        else:
            raise ValueError(f"The file {object_path} was not found")

    def download_from_s3_df(
        self, download_file_name: str, encoding="utf-8"
    ) -> PandasDataFrame:
        df_path = self.download_from_s3(download_file_name)
        return pd.read_csv(df_path, encoding=encoding, low_memory=False)



from typing import Dict, Optional

import pandas as pd
from pandas import DataFrame as PandasDataFrame

from dump.files.tmp_util import TemporaryFileSystem

from .connectors import ConnectorPG, DBUtilsPG


class GetDataManipulatorPG(ConnectorPG):
    def __init__(
        self,
        encoding="utf-16",
        size: int = 8388608,
        db_config_name: str = "postgres",
    ) -> None:
        super().__init__(db_config_name=db_config_name)
        self.encoding = encoding
        self.size = size

        # requires for not execute same query more than one time
        self.__query_history: Dict[str, str] = {}
        self.file_system = TemporaryFileSystem()

    def get_data(self, query: str, path: Optional[str] = None) -> str:
        """
        fetching data from pg using query
        and saving into csv locating in temporary directory
        """
        to_execute = f"""
            copy ( {query} )
            to stdout
            WITH
            CSV HEADER
        """
        if path is None:
            path = self.file_system.save_path()

        if query not in self.__query_history:
            with open(path, "w", encoding=self.encoding) as fp:
                self.cursor.copy_expert(to_execute, fp, size=self.size)
                self.close
                self.__query_history[query] = self.file_system._last_path

        path = self.__query_history[query]
        return path

    def get_df(self, query: str) -> PandasDataFrame:
        path = self.get_data(query)
        return pd.read_csv(path, encoding=self.encoding, low_memory=False)


class SaveDataManipulatorPG(ConnectorPG):
    def __init__(
        self,
        table_name: str,
        db_schema: Optional[str] = None,
        encoding="utf-16",
        size: int = 262144,
        db_config_name: str = "postgres",
    ) -> None:
        super().__init__(db_config_name=db_config_name)
        self.table_name = table_name
        self.db_schema = db_schema
        self.encoding = encoding
        self.size = size

        self.file_system = TemporaryFileSystem()

        self.db_utils = DBUtilsPG(db_config_name)
        if self.db_schema is None:
            self.db_schema = self.db_utils.get_schema_by_table(self.table_name)

    @staticmethod
    def __get_columns_right_order(table_schema: Dict[str, str]) -> list:
        """
        Return list of columns sorted by order in destination table
        """
        column_positions = {
            column: meta_info["position"] for column, meta_info in table_schema.items()
        }
        columns_sorted_by_position = sorted(column_positions, key=column_positions.get)
        return columns_sorted_by_position

    def _validate_df_to_save(self, df) -> PandasDataFrame:
        """
        Check that DataFrame have all required columns
        """
        table_schemа = self.db_utils.get_table_schema(self.table_name, self.db_schema)
        columns_intersection = set(table_schemа).intersection(df.columns)
        columns_not_found = set(table_schemа) - columns_intersection

        if len(columns_not_found) != 0:
            raise Exception(f"Columns not found: {columns_not_found}")

        column_to_write = self.__get_columns_right_order(table_schemа)
        return df[column_to_write]

    def init_df(self, df: PandasDataFrame) -> str:
        """
        Validate df with pg table and save df as csv into temporary directory
        """
        df = self._validate_df_to_save(df)

        path = self.file_system.save_path()
        df.to_csv(path, encoding=self.encoding, index=False)
        return path

    def save_data(
        self,
        path: str,
    ):
        """
        reading file from path
        and saving file to self.table_save

        self.table_save: str - PG table where to save data
        path: str - path to file which should be copied into PG

        Using with combination self.init_data
        or set path manualy
        """

        to_execute = f"""
            copy {self.db_schema}.{self.table_name}
            from STDIN
            WITH
            CSV header
        """

        with open(path, "r", encoding=self.encoding) as fp:
            self.cursor.copy_expert(to_execute, fp, size=self.size)
            self.commit
            self.close

    def save_df(self, df: PandasDataFrame):
        path = self.init_df(df)
        self.save_data(path)
